"""Utilities for working with pydantic-ai types and objects."""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any
from uuid import uuid4

from pydantic_ai import messages as _messages
from pydantic_ai.messages import (
    ModelMessage,
    ModelRequest,
    ModelRequestPart,
    ModelResponse,
    ModelResponsePart,
    RetryPromptPart,
    SystemPromptPart,
    TextPart,
    ToolCallPart,
    ToolReturnPart,
    UserPromptPart,
)

from llmling_agent.messaging.messages import ChatMessage
from llmling_agent.models.content import BaseContent, Content
from llmling_agent.models.tools import ToolCallInfo


if TYPE_CHECKING:
    from llmling_agent.common_types import MessageRole
    from llmling_agent.tools.base import ToolInfo


def format_part(  # noqa: PLR0911
    response: str | _messages.ModelRequestPart | _messages.ModelResponsePart,
) -> str:
    """Format any kind of response in a readable way.

    Args:
        response: Response part to format

    Returns:
        A human-readable string representation
    """
    match response:
        case str():
            return response
        case _messages.TextPart():
            return response.content
        case _messages.ToolCallPart():
            args = str(response.args)
            return f"Tool call: {response.tool_name}\nArgs: {args}"
        case _messages.ToolReturnPart():
            return f"Tool {response.tool_name} returned: {response.content}"
        case _messages.RetryPromptPart():
            if isinstance(response.content, str):
                return f"Retry needed: {response.content}"
            return f"Validation errors:\n{response.content}"
        case _:
            return str(response)


def get_tool_calls(
    messages: list[ModelMessage],
    tools: dict[str, ToolInfo] | None = None,
    agent_name: str | None = None,
    context_data: Any | None = None,
) -> list[ToolCallInfo]:
    """Extract tool call information from messages.

    Args:
        messages: Messages from captured run
        tools: Original ToolInfo set to enrich ToolCallInfos with additional info
        agent_name: Name of the caller
        context_data: Optional context data to attach to tool calls
    """
    tools = tools or {}
    parts = [part for message in messages for part in message.parts]
    call_parts = {
        part.tool_call_id: part
        for part in parts
        if isinstance(part, ToolCallPart) and part.tool_call_id
    }
    return [
        parts_to_tool_call_info(
            call_parts[part.tool_call_id],
            part,
            tools.get(part.tool_name),
            agent_name=agent_name,
            context_data=context_data,
        )
        for part in parts
        if isinstance(part, ToolReturnPart) and part.tool_call_id in call_parts
    ]


def parts_to_tool_call_info(
    call_part: ToolCallPart,
    return_part: ToolReturnPart,
    tool_info: ToolInfo | None,
    agent_name: str | None = None,
    context_data: Any | None = None,
) -> ToolCallInfo:
    """Convert matching tool call and return parts into a ToolCallInfo."""
    args = (
        call_part.args if isinstance(call_part.args, dict) else json.loads(call_part.args)
    )

    return ToolCallInfo(
        tool_name=call_part.tool_name,
        args=args,
        agent_name=agent_name or "UNSET",
        result=return_part.content,
        tool_call_id=call_part.tool_call_id or str(uuid4()),
        timestamp=return_part.timestamp,
        context_data=context_data,
        agent_tool_name=tool_info.agent_name if tool_info else None,
    )


def convert_model_message(  # noqa: PLR0911
    message: ModelMessage | ModelRequestPart | ModelResponsePart,
    tools: dict[str, ToolInfo],
    agent_name: str,
    filter_system_prompts: bool = False,
) -> ChatMessage:
    """Convert a pydantic-ai message to our ChatMessage format.

    Also supports converting parts of a message (with limitations then of course)

    Args:
        message: Message to convert (ModelMessage or its parts)
        tools: Original ToolInfo set to enrich ToolCallInfos with additional info
        agent_name: Name of the agent of this message
        filter_system_prompts: Whether to filter out system prompt parts from
                               ModelRequests

    Returns:
        Converted ChatMessage

    Raises:
        ValueError: If message type is not supported
    """
    match message:
        case ModelRequest():
            # Collect content from all parts
            content_parts = []
            role: MessageRole = "system"
            for part in message.parts:
                match part:
                    case UserPromptPart():
                        content_parts.append(str(part.content))
                        role = "user"
                    case SystemPromptPart() if not filter_system_prompts:
                        content_parts.append(str(part.content))
            return ChatMessage(content="\n".join(content_parts), role=role)

        case ModelResponse():
            # Collect content and tool calls from all parts
            tool_calls = get_tool_calls([message], tools, None)
            parts = [format_part(p) for p in message.parts if isinstance(p, TextPart)]
            content = "\n".join(parts)
            return ChatMessage(content=content, role="assistant", tool_calls=tool_calls)

        case TextPart() as part:
            return ChatMessage(content=format_part(part), role="assistant")

        case UserPromptPart() as part:
            return ChatMessage(content=format_part(part), role="user")

        case SystemPromptPart() as part:
            return ChatMessage(content=format_part(part), role="system")

        case ToolCallPart():
            args = (
                message.args
                if isinstance(message.args, dict)
                else json.loads(message.args)
            )
            info = ToolCallInfo(
                tool_name=message.tool_name,
                args=args,
                agent_name=agent_name,
                result=None,  # Not available yet
                tool_call_id=message.tool_call_id or str(uuid4()),
            )
            content = f"Tool call: {message.tool_name}\nArgs: {args}"
            return ChatMessage(content=content, role="assistant", tool_calls=[info])

        case ToolReturnPart():
            info = ToolCallInfo(
                tool_name=message.tool_name,
                agent_name=agent_name,
                args={},  # No args in return part
                result=message.content,
                tool_call_id=message.tool_call_id or str(uuid4()),
                timestamp=message.timestamp,
            )
            content = f"Tool {message.tool_name} returned: {message.content}"
            return ChatMessage(content=content, role="assistant", tool_calls=[info])

        case RetryPromptPart():
            error_content = (
                message.content
                if isinstance(message.content, str)
                else "\n".join(f"- {err['loc']}: {err['msg']}" for err in message.content)
            )
            return ChatMessage(content=f"Retry needed: {error_content}", role="assistant")

        case _:
            msg = f"Unsupported message type: {type(message)}"
            raise ValueError(msg)


def to_model_message(message: ChatMessage[str | Content]) -> ModelMessage:
    """Convert ChatMessage to pydantic-ai ModelMessage."""
    match message.content:
        case BaseContent():
            content = [message.content.to_openai_format()]
            part = UserPromptPart(content=json.dumps({"content": content}))
            return ModelRequest(parts=[part])
        case str():
            part_cls = {
                "user": UserPromptPart,
                "system": SystemPromptPart,
                "assistant": UserPromptPart,
            }.get(message.role)
            if not part_cls:
                msg = f"Unknown message role: {message.role}"
                raise ValueError(msg)
            return ModelRequest(parts=[part_cls(content=message.content)])



"""Tests for agent configuration models."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import ValidationError
import pytest
import yamling

from llmling_agent import AgentsManifest


if TYPE_CHECKING:
    from pathlib import Path


VALID_AGENT_CONFIG = """\
responses:
  TestResponse:
    description: Test response
    type: inline
    fields:
      message:
        type: str
        description: A message
      score:
        type: int
        constraints:
          ge: 0
          le: 100

agents:
  test_agent:  # Key is the agent ID
    name: Test Agent
    description: A test agent
    model: test
    result_type: TestResponse
    system_prompts:
      - You are a test agent
"""

INVALID_RESPONSE_CONFIG = """\
responses: {}
agent:
  name: Test Agent
  model: test
  result_type: NonExistentResponse
  system_prompts: []
"""


ENV_CONFIG = """\
global_settings:
    llm_capabilities:
        load_resource: false
        get_resources: false
"""

ENV_AGENT = """\
responses:
    BasicResult:
        description: Test result
        type: inline
        fields:
            message:
                type: str
                description: Test message

agents:
    test_agent:
        name: test
        model: test
        result_type: BasicResult
        environment: env.yml  # Relative path
"""


def test_valid_agent_definition():
    """Test valid complete agent configuration."""
    agent_def = AgentsManifest.model_validate(yamling.load_yaml(VALID_AGENT_CONFIG))
    score = agent_def.responses["TestResponse"].fields["score"]  # pyright: ignore
    assert score.constraints == {"ge": 0, "le": 100}


def test_missing_referenced_response():
    """Test referencing non-existent response model."""
    config = yamling.load_yaml(INVALID_RESPONSE_CONFIG)
    with pytest.raises(ValidationError):
        AgentsManifest.model_validate(config)


def test_environment_path_resolution(tmp_path: Path):
    """Test that environment paths are resolved relative to config file."""
    # Create a mock environment config with valid structure
    env_file = tmp_path / "env.yml"
    env_file.write_text(ENV_CONFIG)
    # Create agent config referencing the environment
    config_file = tmp_path / "agents.yml"
    config_file.write_text(ENV_AGENT)

    # Load the config and verify path resolution
    agent_def = AgentsManifest.from_file(config_file)
    test_agent = agent_def.agents["test_agent"]

    # The environment path should now be resolved
    config = test_agent.get_config()
    assert config.global_settings.llm_capabilities.load_resource is False



"""Environment management commands."""

from __future__ import annotations

from typing import TYPE_CHECKING
import webbrowser

from llmling import Config, RuntimeConfig
from slashed import Command, CommandContext, CommandError, PathCompleter

from llmling_agent.models.environment import FileEnvironment, InlineEnvironment


if TYPE_CHECKING:
    from llmling_agent import AgentContext


SET_ENV_HELP = """\
Change the environment configuration file for the current session.

The environment file defines:
- Available tools
- Resource configurations
- Other runtime settings

Example: /set-env configs/new_env.yml

Note: This will reload the runtime configuration and update available tools.
"""

EDIT_ENV_HELP = """\
Open the agent's environment configuration file in the default editor.
This allows you to modify:
- Available tools
- Resources
- Other environment settings
"""


async def set_env(
    ctx: CommandContext[AgentContext],
    args: list[str],
    kwargs: dict[str, str],
):
    """Change the environment file path."""
    from upath import UPath

    if not args:
        await ctx.output.print("Usage: /set-env <path>")
        return

    env_path = args[0]
    if not UPath(env_path).exists():
        msg = f"Environment file not found: {env_path}"
        raise CommandError(msg)

    try:
        agent = ctx.context.agent
        if not agent.context.config:
            msg = "No agent context available"
            raise CommandError(msg)  # noqa: TRY301

        # Manually remove runtime tools
        runtime_tools = [
            name for name, info in agent.tools.items() if info.source == "runtime"
        ]
        for name in runtime_tools:
            del agent.tools[name]

        # Clean up old runtime if we own it
        if agent._owns_runtime and agent.context.runtime:
            await agent.context.runtime.__aexit__(None, None, None)

        # Create and initialize new runtime
        config = Config.from_file(env_path)
        runtime = RuntimeConfig.from_config(config)
        agent.context.runtime = runtime
        agent._owns_runtime = True  # type: ignore

        # Re-initialize agent with new runtime
        await agent.__aenter__()

        await ctx.output.print(
            f"Environment changed to: {env_path}\n"
            f"Replaced runtime tools: {', '.join(runtime_tools)}"
        )

    except Exception as e:
        msg = f"Failed to change environment: {e}"
        raise CommandError(msg) from e


async def edit_env(
    ctx: CommandContext[AgentContext],
    args: list[str],
    kwargs: dict[str, str],
):
    """Open agent's environment file in default application."""
    if not ctx.context.agent.context:
        msg = "No agent context available"
        raise CommandError(msg)

    config = ctx.context.agent.context.config
    match config.environment:
        case FileEnvironment(uri=uri):
            # For file environments, open in browser
            try:
                webbrowser.open(uri)
                await ctx.output.print(f"Opening environment file: {uri}")
            except Exception as e:
                msg = f"Failed to open environment file: {e}"
                raise CommandError(msg) from e
        case InlineEnvironment() as cfg:
            # For inline environments, display the configuration
            yaml_config = cfg.model_dump_yaml()
            await ctx.output.print(f"Inline environment configuration: {yaml_config}")
        case str() as path:
            # Legacy string path
            try:
                resolved = config._resolve_environment_path(path, config.config_file_path)
                webbrowser.open(resolved)
                await ctx.output.print(f"Opening environment file: {resolved}")
            except Exception as e:
                msg = f"Failed to open environment file: {e}"
                raise CommandError(msg) from e
        case None:
            await ctx.output.print("No environment configured")


set_env_cmd = Command(
    name="set-env",
    description="Change the environment configuration file",
    execute_func=set_env,
    usage="<path>",
    help_text=SET_ENV_HELP,
    category="environment",
    completer=PathCompleter(file_patterns=["*.yml", "*.yaml"]),
)

edit_env_cmd = Command(
    name="open-env-file",
    description="Open the agent's environment configuration",
    execute_func=edit_env,
    help_text=EDIT_ENV_HELP,
    category="environment",
    completer=PathCompleter(file_patterns=["*.yml", "*.yaml"]),
)
